import math

from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends.openssl import backend as openssl_backend

from . import Reader


# ====================================================================================================================
# Use this class to read an encrypted PngBin image file.
#
# This class is derived from Reader class, So consult Its documentation for parameter details.
# The exception for the parameters are as below:
#   The multiple of `width` and `height` must be divisible by 4.
#   `key` is a bytes-type and must have a length of 32 bytes.
#   `iv` is a bytes-type and must have a length of 16 bytes, If `offset` >= 16, this parameter can be ignored.
#
# Note:
#   The PngBin file must be encrypted with an AES cipher and CBC mode with key length of 256 bits and iv of 128 bits.
# ====================================================================================================================
class DecryptReader(Reader):
    def __init__(self, width: int, height: int, fobj, key: bytes, iv: bytes, offset: int = 0, length: int = 0):
        """ Creates a reader instance for decrypting an encrypted PngBin image file.

        :param key: a bytes-type and must have a length of 32 bytes.
        :param iv: a bytes-type and must have a length of 16 bytes, If `offset` >= 16, this parameter can be ignored.
        """
        if width * height % 4 != 0:
            raise ValueError('The multiple of `width` and `height` must be divisible by 4.')
        if len(key) != 32:
            raise ValueError('Invalid `key` length (Expected: 32 bytes).')
        if offset < 16 and len(iv) != 16:
            raise ValueError('Invalid `iv` length (Expected: 16 bytes).')
        if not 0 <= offset < width * height * 4:
            raise ValueError('offset must have a value between (0 <= offset < width * height * 4).')

        self.__left = (width * height * 4) - offset  # length in bytes that left to be read.
        if 0 < length < self.__left:
            self.__left = length  # uses `length` if it's in a proper range.

        rem = offset % 16  # A block offset remainder.

        # These two block variables below are always divisible by 16.
        block_length = math.ceil((self.__left + rem) / 16) * 16  # Size in bytes the super class is going to read.
        block_offset = offset - rem  # The offset position to the start of a block.

        # Changes the two block variables for retrieving the initialization vector (IV) from previous block.
        if block_offset != 0:
            block_offset -= 16
            block_length += 16

        # Invokes super class constructor with the appropriate offset and length parameters.
        super().__init__(width, height, fobj, block_offset, block_length)

        _iv = iv if offset - rem == 0 else super().read(16)  # Retrieves the IV from previous block if required.
        self._decryptor = Cipher(algorithms.AES(key), modes.CBC(_iv), backend=openssl_backend).decryptor()

        self._block_buffer = bytearray()  # For storing temporary decrypted bytes from a block.
        self.__left += rem  # Adjusts the state for the next line's read method.
        self.read(rem)  # Reads and discards initial decrypted bytes and stores the rest to temporary buffer.

    @property
    def bytes_left(self) -> int:
        """ length in bytes that left to be read. """
        return self.__left

    def read(self, size: int = -1) -> bytes:
        """ reads fobj, decrypts and returns data that contains inside png.

        :param size: if < 0, reads and returns all the left bytes. otherwise, reads `size` bytes and returns.
        :return: bytes object.
        """
        if not 0 <= size <= self.__left:
            size = self.__left  # if out of range, changes the `size` value to read all that left.
        buffer = bytearray()
        while len(buffer) < size:
            n = size - len(buffer)
            if self._block_buffer:  # Appends bytes that left in the block buffer first.
                n = min(n, len(self._block_buffer))
                buffer.extend(self._block_buffer[:n])
                del self._block_buffer[:n]
            else:
                m = (math.ceil(n / 16) * 16) - n  # A block size complementary (m+n is always divisible by 16).
                blocks = self._decryptor.update(super().read(n + m))  # reads and decrypts in blocks.
                if m > 0:
                    buffer.extend(blocks[:n])  # Appends `n` bytes to `buffer` ...
                    self._block_buffer.extend(blocks[n:])  # ... And puts the rest (`m` bytes) to temp. block buffer.
                else:
                    buffer.extend(blocks)
            self.__left -= n
        return bytes(buffer)



from .Writer import Writer
from .Reader import Reader, InvalidPngError, IncompleteRead
from .EncryptWriter import EncryptWriter
from .DecryptReader import DecryptReader
from .ChainWriter import ChainWriter
from .ChainReader import ChainReader

__all__ = [
    'Writer',
    'Reader',
    'InvalidPngError',
    'IncompleteRead',
    'EncryptWriter',
    'DecryptReader',
    'ChainWriter',
    'ChainReader'
]



import os

from cryptography.hazmat.primitives.ciphers import Cipher, CipherContext, algorithms, modes
from cryptography.hazmat.backends.openssl import backend as openssl_backend

from . import Writer


# ====================================================================================================================
# Use this class to create a PngBin image file with encrypted data.
#
# This class is derived from Writer class, So consult Its documentation for parameter details.
# The exception for the parameters are as below:
#   The multiple of `width` and `height` must be divisible by 4.
#   If `encryptor` is provided, It must be an AES cipher and CBC mode with key length of 256 bits and iv of 128 bits.
# ====================================================================================================================
class EncryptWriter(Writer):
    def __init__(self, width: int, height: int, fobj, auto_finish: bool = False, encryptor: CipherContext = None):
        """ Creates an encrypted PngBin writer instance.

        :param encryptor:
            a CipherContext type. if None, generates a new one with random key and random initialization vector (IV).
        """
        if width * height % 4 != 0:
            raise ValueError('The multiple of `width` and `height` must be divisible by 4.')
        if encryptor:
            if isinstance(encryptor, CipherContext):
                self._encryptor = encryptor
                self._key, self._iv = None, None

                # Access protected class members to gain key and iv info.
                # self._key = encryptor._ctx._cipher.key
                # self._iv = encryptor._ctx._mode.initialization_vector
            else:
                raise TypeError('"encryptor" must be a CipherContext type.')
        else:
            self._encryptor = self._gen_encryptor()

        super().__init__(width, height, fobj, auto_finish)
        self.__left = self._left

    @property
    def key(self) -> bytes:
        """ a cipher key that is generated by this class.
            returns None if you provided an encryptor when creating this class. """
        return self._key

    @property
    def iv(self) -> bytes:
        """ a cipher initialization vector (iv) that is generated by this class.
            returns None if you provided an encryptor when creating this class. """
        return self._iv

    @property
    def bytes_left(self) -> int:
        return self.__left

    def tell(self):
        return self._width * self._height * 4 - self.__left

    def write(self, data: (bytes, bytearray)) -> int:
        """ encrypts and writes input bytes-like object `data` to file.

        :param data: bytes-like object
        :return:
            length in bytes that has been written.
            This is guarantee to be the same as `len(data)` if underlying `fobj.write(data)` is guarantee.
            Unless `bytes_left` is 0 after the call.
        """
        # Developer note:
        #     Do not worry about data left inside `self._encryptor` for the last call,
        #     because when you call `finish()`, all data inside will be written with 0 paddings (if required).

        m = len(data)
        n = min(self.__left, m)
        self.__left -= n

        b = self._encryptor.update(data if m == n else data[:n])
        super().write(b)
        return n

    def _gen_encryptor(self) -> CipherContext:
        """ Generates an encryptor. """
        self._key, self._iv = os.urandom(32), os.urandom(16)
        cipher = Cipher(algorithms.AES(self._key), modes.CBC(self._iv), backend=openssl_backend)
        return cipher.encryptor()



from typing import Iterator

from . import Reader
from . import DecryptReader


# ====================================================================================================================
# Use this class to join multiple PngBin image files and read like they are just a single image.
#
# This class uses either Reader or DecryptReader as a reader class based on `decrypt` parameter.
# This means most of the requirements are going to be the same as those classes.
#
# Note:
#   If you specify `offset` or `length` parameters too large and you don't have enough PngBin file to cover it,
#   It will cause an EOFError exception.
# ====================================================================================================================
class ChainReader:
    def __init__(self, info: Iterator[dict], offset: int, length: int,
                 decrypt: bool = False, auto_close: bool = False):
        """ Creates a reader instance that joins multiple PngBin image files and read like they are just a single image.

        :param info:
            An iterator of dict-type, each item must yield with the only keys of `width`, `height`, and `fobj`
            as the same as expected in Reader and DecryptReader constructor parameters.
            The class will read in order that it yields, from first to last, from end of one file to start of next file.
        :param offset:
            Data-offset of the combined PngBin files.
            This means if `offset` is not within range of the first PngBin file, It will skip and check
            the adjusted offset against the next PngBin file's range in the next iterator and so on.
        :param length: length in bytes expect to read.
        :param decrypt:
            If True, this class also decrypts the data after each read,
            This means each `info` dict needs additional `key` and `iv` keys.
        :param auto_close:
            If True, calls `fobj.close()` on each iteration when skipped, the final byte of each PngBin file is read,
            or `length` bytes have been read (`bytes_left` == 0). It is a no-op if `fobj` does not have `close()`.
        """
        if offset < 0:
            raise ValueError('`offset` cannot be less than 0.')
        if length <= 0:
            raise ValueError('`length` cannot be less than or equal to 0.')

        self._iter_info = info
        self._offset = offset
        self._left = length
        self._reader_cls = DecryptReader if decrypt else Reader
        self._auto_close = auto_close

        while True:  # skips until `offset` is within range.
            self._info = self._get_next_info()
            n = self._info['width'] * self._info['height'] * 4
            if self._offset >= n:
                if self._auto_close and hasattr(self._info['fobj'], 'close'):
                    self._info['fobj'].close()
                self._offset -= n
            else:
                break

        self._reader = self._get_reader()

    @property
    def bytes_left(self) -> int:
        """ length in bytes that left to be read. """
        return self._left

    def read(self, size: int = -1) -> bytes:
        """ reads, decrypts and returns data that contains inside PngBin file.

        :param size:
            If < 0, reads all the left bytes specify by the length.
            Otherwise, reads at specific amount and returns.
        :return: bytes-like object.
        """
        if not 0 <= size <= self._left:
            size = self._left  # if out of range, changes the `size` value to read all that left.
        buffer = bytearray()
        while len(buffer) < size:
            b = self._reader.read(size - len(buffer))
            buffer.extend(b)
            self._left -= len(b)
            if self._left == 0:
                if self._auto_close:
                    self._reader.close()
                break
                # At this point, the next line's `if` statement will always be True,
                # and we don't want to retrieve the next info and create another reader anymore.
            if self._reader.bytes_left == 0:
                if self._auto_close:
                    self._reader.close()
                self._offset = 0  # From now on it always starts at offset 0.
                self._info = self._get_next_info()
                self._reader = self._get_reader()
        return bytes(buffer)

    def close(self):
        self._reader.close()

    def _get_next_info(self) -> dict:
        """ returns the next dict of `info`. """
        try:
            return next(self._iter_info)
        except StopIteration:
            raise EOFError('`info` does not have enough items to read.')

    def _get_reader(self) -> (Reader, DecryptReader):
        """ returns Reader or DecryptReader instance based on `decrypt` constructor parameter. """
        return self._reader_cls(**self._info, offset=self._offset, length=self._left)



import math
import struct
from typing import Tuple, Union


# ====================================================================================================================
# Use this class to read data from a PngBin image file.
# It has an ability to seek to an arbitrary offset of data in the image file.
#
# Terminology:
#   data-offset = A zero-based index offset of the data that contains inside a PngBin file.
#                 Maximum offset is the value of 4 x width x height - 1.
#   png-offset = A zero-based index offset of the whole png file.
#                Maximum offset is the length in bytes of png file subtracts by 1.
# ====================================================================================================================
class Reader:
    def __init__(self, width: int, height: int, fobj, offset: int = 0, length: int = 0):
        """ Creates a PngBin reader instance with offset seeking capability.

        :param width: width of the PngBin image.
        :param height: height of the PngBin image.
        :param fobj:
            if callable:
                It will be called with two png-offset arguments,
                    `first_offset`:
                        The first byte offset this class is going to read.
                    `last_offset` (inclusive):
                        The last byte offset this class is going to read.
                It must return a file-like object that has read(size) method attribute
                with the current read position value of `first_offset`.
            if not callable:
                It has to be a seekable file-like object that has read(size) and seek(pos) method attributes.
        :param offset: data-offset of the PngBin image.
        :param length: length in bytes to read. if length < 1, reads to the end of file.
        """
        if not 0 < width < 2**32:
            raise ValueError('width must have a value between (0 < width < 2**32).')
        if not 0 < height < 2**32:
            raise ValueError('height must have a value between (0 < height < 2**32).')

        if not 0 <= offset < width * height * 4:
            raise ValueError('offset must have a value between (0 <= offset < width * height * 4).')

        self._width, self._height = width, height

        self._left = (self._width * self._height * 4) - offset  # length in bytes that left to be read.
        if 0 < length < self._left:
            self._left = length  # uses `length` if it's in a proper range.

        p_offset, self._nextf, self._nextz = self._convert_offset(self._width, offset, True)
        p_offset_l = self._convert_offset(self._width, (offset + self._left) - 1)  # last png-offset (inclusive).

        self._nextf -= p_offset  # countdown for next filter byte.
        self._nextz -= p_offset  # countdown for next zlib header.

        if callable(fobj):
            self._fobj = fobj(p_offset, p_offset_l)
            if not hasattr(self._fobj, 'read'):
                raise AttributeError('`fobj` must return file-like object that has read method attribute.')
        elif hasattr(fobj, 'read') and hasattr(fobj, 'seek'):
            self._fobj = fobj
            self._fobj.seek(p_offset, 0)
        else:
            raise AttributeError('`fobj` must be callable or file-like object '
                                 'that has read and seek method attributes.')

    @property
    def bytes_left(self) -> int:
        """ length in bytes that left to be read. """
        return self._left

    def read(self, size: int = -1) -> bytes:
        """ reads `fobj` and return data that contains inside png.

        :param size: if < 0, reads and returns all the left bytes. otherwise, reads `size` bytes and returns.
        :return: bytes object.
        """
        if not 0 <= size <= self._left:
            size = self._left  # if out of range, changes the `size` value to read all that left.
        buffer = bytearray()
        while len(buffer) < size:
            n = min(size - len(buffer), self._nextf, self._nextz)
            c = self._fobj.read(n)
            if len(c) != n:
                raise IncompleteRead(
                    f'The length of returning bytes is not equal to what requested. (Expected: {n}, Got: {len(c)})'
                )
            buffer.extend(c)
            self._left -= n   # \
            self._nextf -= n  # -> counting down the 3 variables.
            self._nextz -= n  # /
            if self._left > 0:  # no need to do anything else if there's nothing left.
                if self._nextz == 0:
                    self._read_zlib()
                    self._nextf -= 5  # counting down filter byte.
                    self._nextz = 0xffff  # resets countdown for next zlib header.
                elif self._nextf == 0:
                    if self._fobj.read(1) != b'\x00':
                        raise InvalidPngError('Invalid Filter detected.')
                    self._nextz -= 1  # counting down next zlib header.
                    self._nextf = self._width * 4  # resets countdown for next filter byte.
                    if self._nextf >= self._nextz:  # checks and adds that if zlib header bytes are in the way.
                        self._nextf += math.ceil(self._nextf / 0xffff) * 5
        return bytes(buffer)  # converts to bytes object before returns.

    def close(self):
        """ Calls close() on `fobj` if it has one. """
        if hasattr(self._fobj, 'close'):
            self._fobj.close()

    def _read_zlib(self):
        """ Reads and verifies zlib chunk header for length of 5 bytes.
            if failed, raises InvalidPngError. """
        x, y, z = struct.unpack('<BHH', self._fobj.read(5))
        if x not in [0, 1] or y + z != 0xffff:
            raise InvalidPngError('Invalid zlib detected')

    @staticmethod
    def _convert_offset(width: int, offset: int, extra: bool = False) -> Union[Tuple[int, int, int], int]:
        """ Converts data-offset to png-offset.

        :param width: width of PngBin image.
        :param offset: data-offset of PngBin.
        :param extra: if true, returns tuple of 3 values:
                      (png_offset, next_filter_offset, next_zlib_header_offset)
        :return: if extra is False, returns only converted png-offset
        """
        r = (width * 4)  # number of bytes in a row.
        f = (offset // r) + 1  # number of filter bytes that comes before offset.
        o = offset + f  # offset includes filter bytes.
        c = (o // 0xffff) + 1  # number of zlib chunk headers that comes before offset.
        c *= 5  # each zlib chunk header has length of 5 bytes.
        h = 41 + 2  # fixed header size + zlib main header size.
        p = h + c + o  # converted png-offset.
        if extra:
            nf = r - (offset % r)  # number of bytes from offset until next filter byte.
            nz = 0xffff - (o % 0xffff)  # number of bytes from offset until next zlib header.
            if nf >= nz:  # checks and adds that if zlib header bytes are in the way.
                nf += math.ceil(nf / 0xffff) * 5
            return p, p + nf, p + nz
        else:
            return p


class InvalidPngError(Exception):
    """ Raises when Reader detects an invalid data in a PngBin file.
        This usually means that the image is not created by PngBin. """
    pass


class IncompleteRead(Exception):
    """ Raises when Reader's fobj returns bytes which its length is not equal to what requested. """
    pass



from typing import Iterator, Callable

from . import Writer
from . import EncryptWriter


# ====================================================================================================================
# Use this class to create multiple PngBin image files and write like they are just a single image.
#
# This class uses either Writer or EncryptWriter as a writer class based on `encrypt` parameter.
# This means most of the requirements are going to be the same as those classes.
# One more difference of this class to other 2 writer classes is
# this class will not write anything until the first write(data) method has been called.

# Note: don't forget to call finish() after you're done writing, or use context manager (`with` clause).
# ====================================================================================================================
class ChainWriter:
    def __init__(self, info: Iterator[dict], encrypt: bool = False, on_writer_created: Callable = None):
        """ Creates a writer instance that creates multiple PngBin image files
            and write like they are just a single image.

        :param info:
            An iterator of dict-type, each item must yield with the only keys of `width`, `height`, and `fobj`
            as the same as expected in Writer and EncryptWriter constructor parameters.
            The class will write in order that it yields,
            from first to last, from end of one file to start of next file.
        :param encrypt:
            If True, this class also encrypts the data before each write,
            This means each `info` dict can have an optional `encryptor` key like in EncryptWriter parameter.
        :param on_writer_created:
            If not None and callable, Whenever an underlying writer has been created,
            This callable will be called with that writer instance as an argument.
            This is useful when you want to get `key` and `iv` variables in EncryptWriter instance.
        """
        self._iter_info = info
        self._writer_cls = EncryptWriter if encrypt else Writer
        self._on_writer_created = on_writer_created

        self._is_finished = False
        self._writer = None
        self._wrote = 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.finish()
        
        return False

    @property
    def is_finished(self) -> bool:
        return self._is_finished

    def tell(self):
        return self._wrote

    def write(self, data: (bytes, bytearray)) -> int:
        """ Writes input bytes-like object `data` to file.

        :param data: bytes-like object
        :return:
            length in bytes that has been written.
            This is guaranteed to be the same as `len(data)` if underlying `fobj.write(data)` is guaranteed.
        """
        if self._is_finished:
            raise EOFError('This instance is already finished.')

        n = 0
        while n < len(data):
            if self._writer and self._writer.bytes_left > 0:
                n += self._writer.write(data[n:])
            else:
                self._writer = self._get_writer()
        self._wrote += n
        return n

    def finish(self):
        """ Calls `finish()` on the current writer and makes any future `write(data)` calls to raise an EOFError. """
        if self._writer:
            self._writer.finish()
        self._is_finished = True

    def _get_next_info(self) -> dict:
        """ returns the next dict of `info`. """
        try:
            return next(self._iter_info)
        except StopIteration:
            raise EOFError('`info` does not have enough items to create file.')

    def _get_writer(self) -> (Writer, EncryptWriter):
        """ returns Writer or EncryptWriter instance based on `encrypt` constructor parameter.
            also calls on `on_writer_created` with that writer instance if it is a callable."""
        writer = self._writer_cls(**self._get_next_info(), auto_finish=True)
        if callable(self._on_writer_created):
            self._on_writer_created(writer)
        return writer



import math
import zlib


# ====================================================================================================================
# Use this class to create a PngBin image file and write any data to it.
#
# The result png file format will have only 1 IDAT chunk wth max length of 4,294,967,295 (2**32-1).
# Its color type is RGBA (includes alpha channel).
# The amount of data that can be contained inside png file is the result of 4 x width x height bytes.
# Each zlib chunk has a length of 65,535 (2**16-1) bytes, except for the last chunk which can have arbitrary size.
# No compression nor filter methods are applied.

# Note: don't forget to call finish() after you're done writing, or use context manager (`with` clause).
# Reference: http://www.libpng.org/pub/png/spec/1.2/png-1.2-pdg.html
# ====================================================================================================================
class Writer:
    def __init__(self, width: int, height: int, fobj, auto_finish: bool = False):
        """ Creates a PngBin writer instance.

        :param width: the width of png (0 < width < 2**32).
        :param height: the height of png (0 < height < 2**32).
        :param fobj: file-like object that has write(data) method.
        :param auto_finish: if True, the instance automatically calls finish() when write(data) method
                            detects the last bytes have been written.
        """
        if not 0 < width < 2**32:
            raise ValueError('width must have a value between (0 < width < 2**32).')
        if not 0 < height < 2**32:
            raise ValueError('height must have a value between (0 < height < 2**32).')
        if not hasattr(fobj, 'write'):
            raise AttributeError('fobj must has a write(data) attribute.')

        self._width, self._height = width, height
        self._fobj = fobj
        self._auto_finish = auto_finish

        self._idat_len = self._calc_idat_len(self._width, self._height)
        if self._idat_len >= 2**32:
            raise ValueError('the result of IDAT length cannot be greater than nor equal to 2**32')

        self._left = self._width * self._height * 4  # length in bytes that can be written to.
        self._leftf = self._left + self._height  # same as above but includes filter bytes.
        self._nextf = 0  # countdown number to the next filter byte.
        self._crc32 = 0  # initial value of CRC32.
        self._adler32 = 1  # initial value of Adler32.
        self._is_finished = False  # True if the png footer has already been written.
        self._buffer = bytearray()  # buffer that holds data of each zlib chunk. (max length of 65535)

        self._write_head()  # starts writing header.

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.finish()
        return False

    @staticmethod
    def _calc_idat_len(width: int, height: int) -> int:
        """ Calculates and returns the length in bytes that IDAT is going to have. """
        a = width * height * 4
        b = a + height
        c = b / 0xffff
        d = math.ceil(c)
        e = d * 5
        return 2 + e + b + 4

    @property
    def is_finished(self) -> bool:
        return self._is_finished

    @property
    def bytes_left(self) -> int:
        return self._left

    @property
    def result_length(self) -> int:
        """ Returns the length in bytes of the output png file is going to have. """
        #  header +       IDAT     + footer
        return 41 + self._idat_len + 16

    def tell(self):
        return self._width * self._height * 4 - self._left

    def write(self, data: (bytes, bytearray)) -> int:
        """ Writes input bytes-like object `data` to buffer.
        Note: if there is no space left in this image (`bytes_left` returns 0),
              the rest of this function's call will be no-op and returns 0, unless `auto_finish` is True.

        :param data: bytes-like object
        :return:
            length in bytes that has been written.
            This is guaranteed to be the same as `len(data)` if underlying `fobj.write(data)` is guaranteed,
            unless `bytes_left` is 0 after the call.
        """
        if self._is_finished:
            raise EOFError('`fobj` is already finished.')
        do = 0
        dl = len(data)
        while self._left > 0 and do < dl:
            if len(self._buffer) == 0xffff:
                self._flush_buffer()
            if self._nextf == 0:
                self._buffer.append(0)  # append a filter byte.
                self._nextf = self._width * 4
            n = min(dl - do, 0xffff - len(self._buffer), self._nextf, self._left)
            self._buffer.extend(data[do:do+n])
            do += n
            self._nextf -= n
            self._left -= n
        if self._left == 0 and self._auto_finish:
            self.finish()
        return do

    def finish(self) -> bool:
        """ Flushes remaining buffer and finishes the writing with a footer.
            If there is space left that can be written (`bytes_left` > 0), fills the rest with null-bytes (\x00).
            Note that this method will NOT close the `fobj` for you, call close() instead if you want to do that.

        :return: False if object is already finished, otherwise True.
        """
        if not self._is_finished:
            self._auto_finish = False
            while self._left > 0:
                n = min(2**20, self._left)  # 1MiB max buffer size.
                self.write(bytes(n))
            self._flush_buffer()
            self._write_foot()
            self._is_finished = True
            return True
        else:
            return False

    def close(self):
        """ Calls finish() and then calls close() on `fobj` if it has one. """
        self.finish()
        if hasattr(self._fobj, 'close'):
            self._fobj.close()

    def _flush_buffer(self):
        """ Writes a chunk into fobj. """
        n = len(self._buffer)
        if self._leftf > 0xffff:
            b = b'\x00\xff\xff\x00\x00'  # first byte indicates that this chunk is NOT the last chunk.
        else:
            x = n.to_bytes(2, 'little')  # chunk size in bytes.
            y = (0xffff - n).to_bytes(2, 'little')  # the above is subtracted by 65535.
            b = b'\x01' + x + y  # first byte indicates that this chunk IS the last chunk.
        self._crc32 = zlib.crc32(b, self._crc32)  # updates CRC32.
        self._crc32 = zlib.crc32(self._buffer, self._crc32)
        self._adler32 = zlib.adler32(self._buffer, self._adler32)  # updates Adler32.
        self._fobj.write(b)  # writes to underlying file-object.
        self._fobj.write(self._buffer)  # writes to underlying file-object.
        del self._buffer[:]  # clear the buffer.
        self._leftf -= n

    def _write_head(self):
        self._fobj.write(b'\x89PNG\r\n\x1a\n')  # PNG magic header.
        self._fobj.write(b'\x00\x00\x00\x0d')  # fixed length of IHDR chunk (13).
        b = bytearray()
        b.extend(b'IHDR')
        b.extend(self._width.to_bytes(4, 'big'))  # integer of width in bytes.
        b.extend(self._height.to_bytes(4, 'big'))  # integer of height in bytes.
        b.extend(b'\x08\x06\x00\x00\x00')  # Bit_Depth, Color_Type, Compression_Method, Filter_Method, Interlace_Method
        c = zlib.crc32(b)
        self._fobj.write(b + c.to_bytes(4, 'big'))  # writes data and crc32.
        self._fobj.write(self._idat_len.to_bytes(4, 'big'))  # writes length of idat in bytes.
        b = b'IDATx\x01'  # IDAT name and 2-byte zlib header.
        self._crc32 = zlib.crc32(b, self._crc32)  # updates crc32.
        self._fobj.write(b)

    def _write_foot(self):
        adler32 = self._adler32.to_bytes(4, 'big')
        self._crc32 = zlib.crc32(adler32, self._crc32)
        crc32 = self._crc32.to_bytes(4, 'big')
        footer = b'\x00\x00\x00\x00IEND\xaeB`\x82'
        self._fobj.write(adler32 + crc32 + footer)



import setuptools

with open("README.md", "r", encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    name="PngBin",
    version="0.1.0",
    author="Nathan Young",
    author_email="theyoke3@gmail.com",
    description="Convert any binary data to a PNG image file and vice versa.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TheYoke/PngBin",
    packages=['pngbin'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    license="MIT",
    install_requires=[
        "cryptography",
    ],
)